use std::boxed::Box;
use std::convert::TryInto;
use std::env;

use log::info;
use url::{Host, Url};

use crate::newspaper::Newspaper;
use crate::newspapers::courrier_international::{self, CourrierInternational};
use crate::newspapers::mediapart::{self, Mediapart};
use crate::newspapers::monde_diplomatique::{self, MondeDiplo};

/// Enumerate all errors that can be encountered when using ArticleLocation
#[derive(thiserror::Error, Debug)]
pub enum Error {
    /// The url was not set. Therefore, the article location can't be deduced
    #[error("No url set")]
    NoUrl,
    /// The given URL isn't an accepted Url
    #[error("Malformed URL")]
    MalformedUrl,
    /// The given url doesn't correspond to a newspaper.
    #[error("The given url doesn't link to a known newspaper")]
    UnknownNewspaper,
    /// Error in configuration : used for missing or malformed configuration
    #[error("Error in configuration (configuration key {0} malformed or missing)")]
    Misconfiguration(String),
    /// Other errors
    #[error(transparent)]
    Other(#[from] anyhow::Error),
}

type Newspapers = Vec<Box<dyn Newspaper>>;
pub type Result<T, E = Error> = core::result::Result<T, E>;

fn default_newpapers() -> Result<Newspapers> {
    // TODO: same thing is written too much times : how to DRY ?
    let config_key = "MEDIAPART_COOKIE".to_string();
    let mpruiid = env::var(&config_key)
        .map_err(|_| Error::Misconfiguration(config_key))?
        .into();

    let mediapart = Mediapart::builder()
        .login(mediapart::Login::MPRUUID(mpruiid))
        .build()?;

    let lmd_a_m = "MONDE_DIPLO_LMD_A_M".to_string();
    let phpsessid = "MONDE_DIPLO_PHPSESSID".to_string();
    let spip_session = "MONDE_DIPLO_SPIP_SESSION".to_string();

    let lmd_a_m = env::var(&lmd_a_m)
        .map_err(|_| Error::Misconfiguration(lmd_a_m))?
        .into();
    let phpsessid = env::var(&phpsessid)
        .map_err(|_| Error::Misconfiguration(phpsessid))?
        .into();
    let spip_session = env::var(&spip_session)
        .map_err(|_| Error::Misconfiguration(spip_session))?
        .into();

    let monde_diplo = MondeDiplo::builder()
        .login(monde_diplomatique::Login::Cookies {
            lmd_a_m,
            phpsessid,
            spip_session,
        })
        .build()?;

    let lmd_a_m = "COURRIER_INTERNATIONAL_LMD_A_M".to_string();
    let ssess = "COURRIER_INTERNATIONAL_SSESS".to_string();

    let lmd_a_m = env::var(&lmd_a_m)
        .map_err(|_| Error::Misconfiguration(lmd_a_m))?
        .into();
    let ssess = env::var(&ssess)
        .map_err(|_| Error::Misconfiguration(ssess))?
        .into();

    let courrier_international = CourrierInternational::builder()
        .login(courrier_international::Login::Cookies { lmd_a_m, ssess })
        .build()?;

    Ok(vec![
        Box::new(mediapart),
        Box::new(monde_diplo),
        Box::new(courrier_international),
    ])
}

#[derive(Default)]
pub struct Builder {
    url: Option<Url>,
    newspapers: Option<Newspapers>,
}

impl Builder {
    pub fn new() -> Self {
        Self::default()
    }

    /// Adds an url corresponding to article location
    ///
    /// # Errors
    ///
    /// An error is returned if the could not be converted into an url
    pub fn url<U, E>(mut self, url: U) -> Result<Self>
    where
        U: TryInto<Url, Error = E> + Send,
        E: std::error::Error + Sync + Send + 'static,
    {
        let url = url.try_into().map_err(|_| Error::MalformedUrl)?;
        self.url = Some(url);
        Ok(self)
    }

    /// Adds a newspaper to the list
    pub fn newspaper<T>(mut self, newspaper: T) -> Self
    where
        T: 'static + Newspaper,
    {
        match &mut self.newspapers {
            Some(newspapers) => newspapers.push(Box::new(newspaper)),
            None => self.newspapers = Some(vec![Box::new(newspaper)]),
        }
        self
    }

    /// Adds several newspapers to the list of accepted newspapers
    pub fn newspapers(mut self, newspapers: Newspapers) -> Self {
        match &mut self.newspapers {
            Some(current_newspapers) => current_newspapers.extend(newspapers),
            None => self.newspapers = Some(newspapers.into_iter().collect::<Vec<_>>()),
        };
        self
    }

    /// Builds the ArticleLocation by looking which newspaper
    ///
    /// # Errors
    ///
    /// The following errors can be returned
    ///
    /// - no newpspaper is given
    /// - the url is not set
    /// - the given url has no host
    pub fn build(self) -> Result<ArticleLocation> {
        let url = Clone::clone(self.url.as_ref().ok_or(Error::NoUrl)?);
        let host = url.host_str().ok_or(Error::MalformedUrl)?;
        let host = Host::parse(host).map_err(|_| Error::MalformedUrl)?;
        let newspaper = self
            .newspapers
            .unwrap_or(default_newpapers()?)
            .into_iter()
            .find(|c| c.metadata().hosts.contains(&host))
            .ok_or(Error::UnknownNewspaper)?;
        Ok(ArticleLocation { newspaper, url })
    }
}

pub struct ArticleLocation {
    newspaper: Box<dyn Newspaper>,
    pub url: Url,
}

impl ArticleLocation {
    pub fn builder() -> Builder {
        Builder::new()
    }

    pub async fn retrieve_html(&self) -> Result<String> {
        // TODO: modify when retrieve_html returns a specific Error type
        Ok(self.newspaper.retrieve_html(&self.url).await?)
    }
}
